with Reps.Ops, Rounding;

pragma Elaborate_All (Reps.Ops);

package body Intervals is

   NegOne:     constant Rep := -One;
   NegTwo:     constant Rep := -Two;
   NegHalf:    constant Rep := -Half;
   Three:      constant Rep := Rep(3);
   Quarter:    constant Rep := Half/Two;
   OnePlus:    constant Rep := Rep(33)/Rep(32);
   Epsilon:    constant Rep := Half**99;
   Deltaa:    constant Rep := Half**50;

   Interval_Zero: constant Interval := (Zero,Zero,Zero);
   Interval_One:  constant Interval := (One,Zero,Zero);
   Interval_Two:  constant Interval := (Two,Zero,Zero);
   PlusMinus: constant Interval := (Zero,One,Zero);

   function IsNumeric(S: Interval) return Boolean is
   begin
      return False;
   end IsNumeric;


   function IsNumbers(S: Interval) return Boolean is
   begin
      return (S.E=Zero);
   end IsNumbers;


   function Err0(S: Interval) return Boolean is
   begin
      return S.R=Zero and S.E=Zero;
   end Err0;
   
   
   function Scal(K: Integer) return Interval is
   begin
      return (Rep(K),Zero,Zero);
   end Scal;
   
   
   function Scal(R: Rep) return Interval is
   begin
      return (R,Zero,Zero);
   end Scal;
   
   
   function Approx(S: Interval) return Rep is
   begin
      return S.C;
   end Approx;
   
   
   procedure ModNumbers(S: in out Interval) is
   begin
      S.C := Zero;
      S.R := Zero;
   end ModNumbers;


   function Ball(R: Rep) return Interval is
   begin
      return (Zero,Abs(R),Zero);
   end Ball;
   
   
   function Ball(R,E: Rep) return Interval is
   begin
      return (Zero,Abs(R),Abs(E));
   end Ball;
   
   
   function Ball(S: Interval) return Interval is
   begin
      return (Zero,Abs(S.C)+S.R,Zero);
   end Ball;


   function Disk(S: Interval) return Interval is
   begin
      return (S.C,S.R+S.E,Zero);
   end Disk;


   function MaxAbs(S: Interval) return Radius is
   begin
      return (abs(S.C)+S.R)+S.E;
   end MaxAbs;


   function Scal(R1,R2: Rep) return Interval is
      Sum: constant Rep := R1+R2;
   begin
      if R1 >= R2 then
	 return (Half*Sum,Half*((R1-R2)+(Sum+(-R1-R2))),Zero);
      end if;
      return (Half*Sum,Half*((R2-R1)+(Sum+(-R1-R2))),Zero);
   end Scal;
   
   
   function Center(S: Interval) return Interval is
   begin
      return (S.C,Zero,Zero);
   end Center;
   
   
   function Contains(S: Interval; Z: Rep) return Boolean is
   begin
      return (Sup(S)>=Z and Inf(S)<=Z);
   end Contains;
   
   
   procedure ResetCenter(S: in out Interval) is
   begin
      S.C := Zero;
   end ResetCenter;
   
   
   function ResetCenter(S: Interval) return Interval is
   begin
      return (Zero,S.R,S.E);
   end ResetCenter;
   
   
   function ResetBall(S: Interval) return Interval is
   begin
      return (Zero,S.E,Zero);
   end ResetBall;


   procedure Split(S: in Interval; S0,SE: out Interval) is
   begin
      S0.C := S.C;
      S0.R := Zero;
      S0.E := Zero;
      SE.C := Zero;
      SE.R := S.R;
      SE.E := S.E;
   end Split;
   
   
   function Widen(S: Interval) return Interval is
      -- only used in high level procedures
   begin
      return (S.C,S.R+Deltaa,S.E+Deltaa); -- avoid Gnat's Succ in Round_Up mode
   end Widen;
   
   
   function Inf(S: Interval) return Rep is
   begin
      return -(S.R-S.C);
   end Inf;
   
   
   function Sup(S: Interval) return Rep is
   begin
      return S.C+S.R;
   end Sup;
   
   
   pragma Inline (Inf,Sup);
   
   
   function Union(S1,S2: Interval) return Interval is
   begin
      return Scal(Rep'Min(Inf(S1),Inf(S2)),Rep'Max(Sup(S1),Sup(S2)));
   end Union;
   
   
   function "<"(S1,S2: Interval) return Boolean is
   begin
      if Sup(S1) < Inf(S2) then return True; end if;
      if Inf(S1) < Sup(S2) then raise Undefined; end if;
      return False;
   end "<";
   
   
   function "<="(S1,S2: Interval) return Boolean is
   begin
      if Sup(S1) <= Inf(S2) then return True; end if;
      if Inf(S1) <= Sup(S2) then raise Undefined; end if;
      return False;
   end "<=";
   
   
   function ">"(S1,S2: Interval) return Boolean is
   begin
      if Inf(S1) > Sup(S2) then return True; end if;
      if Sup(S1) > Inf(S2) then return True; end if;
      return False;
   end ">";
   
   
   function ">="(S1,S2: Interval) return Boolean is
   begin
      if Inf(S1) >= Sup(S2) then return True; end if;
      if Sup(S1) >= Inf(S2) then return True; end if;
      return False;
   end ">=";
   
   
   function Min(S1,S2: Interval) return Interval is
      U1: constant Rep := Sup(S1);
      L2: constant Rep := Inf(S2);
      L1,U2: Rep;
   begin
      if U1 <= L2 then
	 return S1;
      end if;
      L1 := Inf(S1);
      U2 := Sup(S2);
      if U2 <= L1 then
	 return S2;
      end if;
      return Scal(Rep'Min(L1,L2),Rep'Min(U1,U2));
   end Min;
   
   
   function Max(S1,S2: Interval) return Interval is
      U1: constant Rep := Sup(S1);
      L2: constant Rep := Inf(S2);
      L1,U2: Rep;
   begin
      if U1 <= L2 then
	 return S2;
      end if;
      L1 := Inf(S1);
      U2 := Sup(S2);
      if U2 <= L1 then
	 return S1;
      end if;
      return Scal(Rep'Max(L1,L2),Rep'Max(U1,U2));
   end Max;


   function Norm(S: Interval) return Interval is
      T: Interval;
   begin
      T.C := abs(S.C);
      T.R := S.R+S.E;
      T.E := Zero;
      if T.R>T.C then
	 T.C := Half*(T.C+T.R);
	 T.R := T.C;
      end if;
      return T;
   end Norm;


   function InfAbs(S: Interval) return Rep is
   begin
      return -Rep'Min(S.R-Abs(S.C),Zero);
   end InfAbs;
   
   
   function SupAbs(S: Interval) return Rep is
   begin
      return Abs(S.C)+S.R;
   end SupAbs;
   
   
   function "-"(S: Interval) return Interval is
   begin
      return (-S.C,S.R,S.E);
   end "-";
   
   
   function "+"(S,T: Interval) return Interval is
      Sum: constant Rep := S.C+T.C;
   begin
      return (Sum,S.R+T.R+(Sum+(-S.C-T.C)), S.E+T.E );
   end "+";
   
   
   function "-"(S,T: Interval) return Interval is
      Diff: constant Rep := S.C-T.C;
   begin
      return (Diff,S.R+T.R+(Diff+(T.C-S.C)), S.E+T.E);
   end "-";
   
   
   function "*"(R: Rep; S: Interval) return Interval is
      Prod: constant Rep := R*S.C;
   begin
      return (Prod,Abs(R)*S.R+(Prod+(-R)*S.C), Abs(R)*S.E );
   end "*";
   
   
   procedure ErrMult(R: in Radius; S: in out Interval) is
   begin
      S.R := R*S.R;
      S.E := R*S.E;
   end ErrMult;
   
   
   function "/"(S: Interval; R: Rep) return Interval is
      Quot: constant Rep := S.C/R;
   begin
      return (Quot,S.R/Abs(R)+(Quot+S.C/(-R)), S.E/Abs(R)  );
   end "/";
   
   
   function Sqr(S: Interval) return Interval is
      Prod: constant Rep := S.C*S.C;
   begin
      return (Prod,(Two*Abs(S.C)+S.R)*S.R+(Prod+S.C*(-S.C)), (Two*(Abs(S.C)+S.R)+S.E)*S.E);
   end Sqr;
   
   
   function "*"(S,T: Interval) return Interval is
      Prod: constant Rep := S.C*T.C;
   begin
      return (Prod, Abs(S.C)*T.R+S.R*(Abs(T.C)+T.R)+(Prod+S.C*(-T.C)), Abs(S.C)*T.E+S.E*(Abs(T.C)+T.E));
   end "*";
   
   
   procedure MultAdd(S1,S2: in Interval; S3: in out Interval) is
      T: Interval;
   begin
      T:=S1*S2;
      S3:=S3+T;
   end MultAdd;


   function Inv(S: Interval) return Interval is
      Quot: constant Rep := One/S.C;
   begin
      if Abs(S.C) <= S.R then
	 raise Constraint_Error;               -- division by zero
      end if;
      --     return (Quot,(S.R/Abs(S.C))/(-(S.R-Abs(S.C)))+(Quot+NegOne/S.C), S.E/Abs(S.C)/(-(S.E-Abs(S.C)))  );
      return (Quot,(S.R/Abs(S.C))/(-(S.R-Abs(S.C)))+(Quot+NegOne/S.C), S.E );
   end Inv;
   
   
   function "/"(S,T: Interval) return Interval is
   begin
      return S*Inv(T);
   end "/";
   
   
   function "**"(S: Interval; I: Integer) return Interval is
   begin
      if I>0 then
	 declare
	    J: Integer := I;
	    X,Y: Interval;
	 begin
	    X := S;
	    if (J rem 2)=0 then
	       Y := Interval_One;
	    else
	       Y := X;
	    end if;
	    J := J/2;
	    while J>0 loop
	       X := Sqr(X);
	       if (J rem 2)>0 then
		  Y := X*Y;
	       end if;
	       J := J/2;
	    end loop;
	    return Y;
	 end;
      elsif I<0 then
	 return Inv(S)**(-I);
      else
	 return Interval_One;
      end if;
   end "**";
   
   
   procedure CoshSinh(S: in Interval; Sc,Ss: out Interval) is
      Iter: constant Integer := 6;
   begin
      if S = Interval_Zero then
	 Sc := Interval_One;
	 Ss := S;
      else
	 declare
	    K: Integer;
	    R: Rep;
	    X,T: Interval;
	 begin
	    K := 0;
	    R := SupAbs(S);
	    while R>Half loop
	       K := K+1;
	       R := R*Half;
	    end loop;
	    X := (Half**K)*S;
	    T := Sqr(X);
	    Sc := (One,Half,Zero);
	    Ss := Sc;
	    for I in reverse 1 .. Iter loop
	       Sc := Interval_One+Sc*T/Rep((2*I)*(2*I-1));
	       Ss := Interval_One+Ss*T/Rep((2*I)*(2*I+1));
	    end loop;
	    Ss := X*Ss;
	    for I in 1 .. K loop
	       T := Sc;
	       Sc := Interval_One+Two*Sqr(Ss);
	       Ss := Two*(T*Ss);
	    end loop;
	 end;
      end if;
   end CoshSinh;
   
   
   function Cosh(S: Interval) return Interval is
      Sc,Ss: Interval;
   begin
      CoshSinh(S,Sc,Ss);
      return Sc;
   end Cosh;
   
   
   function Sinh(S: Interval) return Interval is
      Sc,Ss: Interval;
   begin
      CoshSinh(S,Sc,Ss);
      return Ss;
   end Sinh;


   procedure CosSin(S: in Interval; Sc,Ss: out Interval) is
      Iter: constant Integer := 8;
      Interval_Pi: constant Interval:=Scal(Rep(3.14159_26535_89793_23),Rep(3.14159_26535_89793_24));
      Co: Interval;
   begin
      if S = Interval_Zero then
         Sc := Interval_One;
         Ss := S;
      else
         declare
            K: Integer;
            R: Rep;
            X,T: Interval;
         begin
            K := 0;
            R := SupAbs(S);
            while R>Half loop
               K := K+1;
               R := R*Half;
            end loop;
            X := (Half**K)*S;
            T := Sqr(X);
            Sc := (One,Half,Zero);
            Ss := Sc;
            if ((Iter-1) mod 2)=0 then
               Co:=Interval_One;
            else
               Co:=-Interval_One;
            end if;
            for I in reverse 1 .. Iter loop
               Sc := Co+Sc*T/Rep((2*I)*(2*I-1));
               Ss := Co+Ss*T/Rep((2*I)*(2*I+1));
               Co:=-Co;
            end loop;

            Ss := X*Ss;
            for I in 1 .. K loop
               T := Sc;
               Sc := Interval_One-Two*Sqr(Ss);
               Ss := Two*(T*Ss);
            end loop;
         end;
      end if;
   end CosSin;


   function Cos(S: Interval) return Interval is
      Sc,Ss: Interval;
   begin
      CosSin(S,Sc,Ss);
      return Sc;
   end Cos;
   
   
   function Sin(S: Interval) return Interval is
      Sc,Ss: Interval;
   begin
      CosSin(S,Sc,Ss);
      return Ss;
   end Sin;


   function ExpUp(R: Rep) return Rep is
      Iter: constant Integer := 13;
      K: Integer := 0;
      X: Rep := R;
      Y: Rep := OnePlus;         --  33/32  assumes Iter>8
   begin
      if R<Zero then return One/ExpDown(-R); end if;
      if R=Zero then return One; end if;
      while X>Quarter loop
	 K := K+1;
	 X := Half*X;
      end loop;
      for I in reverse 1 .. Iter loop
	 Y := One+X*Y/Rep(I);
      end loop;
      for I in 1 .. K loop
	 Y := Y*Y;
      end loop;
      return Y;
   end ExpUp;
   
   
   function ExpDown(R: Rep) return Rep is
      Iter: constant Integer := 13;
      K: Integer := 0;
      X: Rep := R;
      Y: Rep := NegOne;
   begin
      if R<Zero then return -(NegOne/ExpUp(-R)); end if;
      if R=Zero then return One; end if;
      while X>Quarter loop
	 K := K+1;
	 X := Half*X;
      end loop;
      for I in reverse 1 .. Iter loop
	 Y := NegOne+X*Y/Rep(I);
      end loop;
      for I in 1 .. K loop
	 Y := (-Y)*Y;
      end loop;
      return -Y;
   end ExpDown;
   
   
   function Exp(S: Interval) return Interval is
   begin
      return Scal(ExpDown(Inf(S)),ExpUp(Sup(S)));
   end Exp;
   
   
   function Log(S: Interval) return Interval is
      use Reps.Ops;
   begin
      if S=Interval_One then return Interval_Zero; end if;
      declare
	 Iter: constant Integer := 3;
	 S0,LogS0,X,Y: Interval;
      begin
	 LogS0 := Scal(Log(Approx(S)));   --- guess
	 S0 := Exp(LogS0);
	 X := (S0-S)/S0;                  --- Log(S) = Log(S0)+Log(1-X)
	 Y := Interval_One-Scal(SupAbs(X));
	 if Inf(Y) <= Zero then raise Constraint_Error; end if;
	 Y := PlusMinus*Inv(Y)/Rep(Iter+1);
	 for I in reverse 1 .. Iter loop
	    Y := Inv(Scal(I))+X*Y;
	 end loop;
	 return LogS0-X*Y;
      end;
   end Log;
   
   
   function Sqrt(S: Interval) return Interval is
      use Reps.Ops;
   begin
      if S=Interval_Zero then return Interval_Zero; end if;
      if Inf(S)<Zero then raise Undefined; end if;
      declare
	 S0,SqrtS0,X,Y: Interval;
      begin
	 SqrtS0 := Scal(Sqrt(Approx(S))); --- guess
	 S0 := Sqr(SqrtS0);
	 X := (S-S0)/S0;                  --- Sqrt(S) = Sqrt(S0)*Sqrt(1+X)
	 Y := (((Zero,One,Zero)*X-Scal(2))*X+Scal(8))*X/Scal(16);
	 return SqrtS0*(Interval_One+Y);
      end;
   end Sqrt;
   
   
   function TruncExp(N: Natural; S: Interval) return Interval is
      Y: Interval := Interval_One;
   begin
      for I in reverse 1 .. N loop
	 Y := Interval_One+(S*Y)/Rep(I);
      end loop;
      return Y;
   end TruncExp;
   
   
   function RemExp(N: Natural; S: Interval) return Interval is
      --- Taylor remainder for Exp, divided by S**(N+1)/(N+1)!
      M: Integer;
      R: constant Rep := Sup(S);
      Y: Interval;
   begin
      if Inf(S)<Zero then
	 raise Undefined;
      elsif R<Half then
	 M := Integer'Max(N+1,17);
      elsif R<Two then
	 M := Integer'Max(N+1,26);
      elsif R<Rep(8) then
	 M := Integer'Max(N+1,46);
      elsif R<Rep(32) then
	 M := Integer'Max(N+1,96);
      elsif R<Rep(128) then
	 M := Integer'Max(N+1,245);
      elsif R<Rep(300) then
	 M := Integer'Max(N+1,470);
      elsif R<Rep(512) then
	 M := Integer'Max(N+1,730);
      elsif R<Rep(2048) then
	 M := Integer'Max(N+1,1500);
      else
	 raise Undefined;
      end if;
      Y := Inv(Interval_One-Scal(Zero,R)/Rep(M+1));
      for I in reverse N+2 .. M loop
	 Y := Interval_One+(S*Y)/Rep(I);
      end loop;
      return Y;
   end RemExp;

   begin
      Rounding.Restore_Checks;

end Intervals;
